;; Machine Description for LARCH Loongson SX ASE
;;
;; Copyright (C) 2018 Free Software Foundation, Inc.
;;
;; This file is part of GCC.
;;
;; GCC is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 3, or (at your option)
;; any later version.
;;
;; GCC is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with GCC; see the file COPYING3.  If not see
;; <http://www.gnu.org/licenses/>.
;;

(define_mode_attr d2lsxfmt
  [(V4SI "q")
   (V8HI "d")
   (V16QI "w")])

(define_mode_attr d2lsxfmt_u
  [(V4SI "qu")
   (V8HI "du")
   (V16QI "wu")])

;; The attribute gives two double modes for vector modes.
(define_mode_attr VD2MODE
  [(V4SI "V2DI")
   (V8HI "V2DI")
   (V16QI "V4SI")])

(define_c_enum "unspec" [
  UNSPEC_LSX_VADDWEV
  UNSPEC_LSX_VADDWEV2
  UNSPEC_LSX_VADDWEV3
  UNSPEC_LSX_VADDWOD
  UNSPEC_LSX_VADDWOD2
  UNSPEC_LSX_VADDWOD3
  UNSPEC_LSX_VSUBWEV
  UNSPEC_LSX_VSUBWEV2
  UNSPEC_LSX_VSUBWOD
  UNSPEC_LSX_VSUBWOD2
  UNSPEC_LSX_VMULWEV
  UNSPEC_LSX_VMULWEV2
  UNSPEC_LSX_VMULWEV3
  UNSPEC_LSX_VMULWOD
  UNSPEC_LSX_VMULWOD2
  UNSPEC_LSX_VMULWOD3
  UNSPEC_LSX_VHADDW_Q_D
  UNSPEC_LSX_VHADDW_QU_DU
  UNSPEC_LSX_VHSUBW_Q_D
  UNSPEC_LSX_VHSUBW_QU_DU
  UNSPEC_LSX_VMADDWEV
  UNSPEC_LSX_VMADDWEV2
  UNSPEC_LSX_VMADDWEV3
  UNSPEC_LSX_VMADDWOD
  UNSPEC_LSX_VMADDWOD2
  UNSPEC_LSX_VMADDWOD3
  UNSPEC_LSX_VROTR
  UNSPEC_LSX_VADD_Q
  UNSPEC_LSX_VSUB_Q
  UNSPEC_LSX_VEXTH_Q_D 
  UNSPEC_LSX_VEXTH_QU_DU 
  UNSPEC_LSX_VMSKGEZ
  UNSPEC_LSX_VMSKNZ
  UNSPEC_LSX_VROTRI
  UNSPEC_LSX_VEXTL_Q_D 
  UNSPEC_LSX_VSRLNI
  UNSPEC_LSX_VSRLRNI
  UNSPEC_LSX_VSSRLNI
  UNSPEC_LSX_VSSRLNI2
  UNSPEC_LSX_VSSRLRNI
  UNSPEC_LSX_VSSRLRNI2
  UNSPEC_LSX_VSRANI
  UNSPEC_LSX_VSRARNI
  UNSPEC_LSX_VSSRANI
  UNSPEC_LSX_VSSRANI2
  UNSPEC_LSX_VSSRARNI
  UNSPEC_LSX_VSSRARNI2
  UNSPEC_LSX_VPERMI
])

(define_insn "lsx_v<optab>wev_d_w<u>"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(addsubmul:V2DI
	  (any_extend:V2DI
	    (vec_select:V2SI
	      (match_operand:V4SI 1 "register_operand" "%f")
	      (parallel [(const_int 0) (const_int 2)])))
	  (any_extend:V2DI
	    (vec_select:V2SI
	      (match_operand:V4SI 2 "register_operand" "f")
	      (parallel [(const_int 0) (const_int 2)])))))]
  "ISA_HAS_LSX"
  "v<optab>wev.d.w<u>\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_v<optab>wev_w_h<u>"
  [(set (match_operand:V4SI 0 "register_operand" "=f")
	(addsubmul:V4SI
	  (any_extend:V4SI
	    (vec_select:V4HI
	      (match_operand:V8HI 1 "register_operand" "%f")
	      (parallel [(const_int 0) (const_int 2)
	        	 (const_int 4) (const_int 6)])))
	  (any_extend:V4SI
	    (vec_select:V4HI
	      (match_operand:V8HI 2 "register_operand" "f")
	      (parallel [(const_int 0) (const_int 2)
			 (const_int 4) (const_int 6)])))))]
  "ISA_HAS_LSX"
  "v<optab>wev.w.h<u>\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V4SI")])

(define_insn "lsx_v<optab>wev_h_b<u>"
  [(set (match_operand:V8HI 0 "register_operand" "=f")
	(addsubmul:V8HI
	  (any_extend:V8HI
	    (vec_select:V8QI
	      (match_operand:V16QI 1 "register_operand" "%f")
	      (parallel [(const_int 0) (const_int 2)
			 (const_int 4) (const_int 6)
			 (const_int 8) (const_int 10)
			 (const_int 12) (const_int 14)])))
	  (any_extend:V8HI
	    (vec_select:V8QI
              (match_operand:V16QI 2 "register_operand" "f")
	      (parallel [(const_int 0) (const_int 2)
			 (const_int 4) (const_int 6)
			 (const_int 8) (const_int 10)
			 (const_int 12) (const_int 14)])))))]
  "ISA_HAS_LSX"
  "v<optab>wev.h.b<u>\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V8HI")])

(define_insn "lsx_v<optab>wod_d_w<u>"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(addsubmul:V2DI
	  (any_extend:V2DI
	    (vec_select:V2SI
	      (match_operand:V4SI 1 "register_operand" "%f")
	      (parallel [(const_int 1) (const_int 3)])))
	  (any_extend:V2DI
	    (vec_select:V2SI
	      (match_operand:V4SI 2 "register_operand" "f")
	      (parallel [(const_int 1) (const_int 3)])))))]
  "ISA_HAS_LSX"
  "v<optab>wod.d.w<u>\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_v<optab>wod_w_h<u>"
  [(set (match_operand:V4SI 0 "register_operand" "=f")
	(addsubmul:V4SI
	  (any_extend:V4SI
	    (vec_select:V4HI
	      (match_operand:V8HI 1 "register_operand" "%f")
	      (parallel [(const_int 1) (const_int 3)
	        	 (const_int 5) (const_int 7)])))
	  (any_extend:V4SI
	    (vec_select:V4HI
	      (match_operand:V8HI 2 "register_operand" "f")
	      (parallel [(const_int 1) (const_int 3)
			 (const_int 5) (const_int 7)])))))]
  "ISA_HAS_LSX"
  "v<optab>wod.w.h<u>\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V4SI")])

(define_insn "lsx_v<optab>wod_h_b<u>"
  [(set (match_operand:V8HI 0 "register_operand" "=f")
	(addsubmul:V8HI
	  (any_extend:V8HI
	    (vec_select:V8QI
	      (match_operand:V16QI 1 "register_operand" "%f")
	      (parallel [(const_int 1) (const_int 3)
			 (const_int 5) (const_int 7)
			 (const_int 9) (const_int 11)
			 (const_int 13) (const_int 15)])))
	  (any_extend:V8HI
	    (vec_select:V8QI
              (match_operand:V16QI 2 "register_operand" "f")
	      (parallel [(const_int 1) (const_int 3)
			 (const_int 5) (const_int 7)
			 (const_int 9) (const_int 11)
			 (const_int 13) (const_int 15)])))))]
  "ISA_HAS_LSX"
  "v<optab>wod.h.b<u>\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V8HI")])

(define_insn "lsx_v<optab>wev_d_wu_w"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(addmul:V2DI
	  (zero_extend:V2DI
	    (vec_select:V2SI
	      (match_operand:V4SI 1 "register_operand" "%f")
	      (parallel [(const_int 0) (const_int 2)])))
	  (sign_extend:V2DI
	    (vec_select:V2SI
	      (match_operand:V4SI 2 "register_operand" "f")
	      (parallel [(const_int 0) (const_int 2)])))))]
  "ISA_HAS_LSX"
  "v<optab>wev.d.wu.w\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_v<optab>wev_w_hu_h"
  [(set (match_operand:V4SI 0 "register_operand" "=f")
	(addmul:V4SI
	  (zero_extend:V4SI
	    (vec_select:V4HI
	      (match_operand:V8HI 1 "register_operand" "%f")
	      (parallel [(const_int 0) (const_int 2)
	        	 (const_int 4) (const_int 6)])))
	  (sign_extend:V4SI
	    (vec_select:V4HI
	      (match_operand:V8HI 2 "register_operand" "f")
	      (parallel [(const_int 0) (const_int 2)
			 (const_int 4) (const_int 6)])))))]
  "ISA_HAS_LSX"
  "v<optab>wev.w.hu.h\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V4SI")])

(define_insn "lsx_v<optab>wev_h_bu_b"
  [(set (match_operand:V8HI 0 "register_operand" "=f")
	(addmul:V8HI
	  (zero_extend:V8HI
	    (vec_select:V8QI
	      (match_operand:V16QI 1 "register_operand" "%f")
	      (parallel [(const_int 0) (const_int 2)
			 (const_int 4) (const_int 6)
			 (const_int 8) (const_int 10)
			 (const_int 12) (const_int 14)])))
	  (sign_extend:V8HI
	    (vec_select:V8QI
              (match_operand:V16QI 2 "register_operand" "f")
	      (parallel [(const_int 0) (const_int 2)
			 (const_int 4) (const_int 6)
			 (const_int 8) (const_int 10)
			 (const_int 12) (const_int 14)])))))]
  "ISA_HAS_LSX"
  "v<optab>wev.h.bu.b\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V8HI")])

(define_insn "lsx_v<optab>wod_d_wu_w"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(addmul:V2DI
	  (zero_extend:V2DI
	    (vec_select:V2SI
	      (match_operand:V4SI 1 "register_operand" "%f")
	      (parallel [(const_int 1) (const_int 3)])))
	  (sign_extend:V2DI
	    (vec_select:V2SI
	      (match_operand:V4SI 2 "register_operand" "f")
	      (parallel [(const_int 1) (const_int 3)])))))]
  "ISA_HAS_LSX"
  "v<optab>wod.d.wu.w\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_v<optab>wod_w_hu_h"
  [(set (match_operand:V4SI 0 "register_operand" "=f")
	(addmul:V4SI
	  (zero_extend:V4SI
	    (vec_select:V4HI
	      (match_operand:V8HI 1 "register_operand" "%f")
	      (parallel [(const_int 1) (const_int 3)
	        	 (const_int 5) (const_int 7)])))
	  (sign_extend:V4SI
	    (vec_select:V4HI
	      (match_operand:V8HI 2 "register_operand" "f")
	      (parallel [(const_int 1) (const_int 3)
			 (const_int 5) (const_int 7)])))))]
  "ISA_HAS_LSX"
  "v<optab>wod.w.hu.h\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V4SI")])

(define_insn "lsx_v<optab>wod_h_bu_b"
  [(set (match_operand:V8HI 0 "register_operand" "=f")
	(addmul:V8HI
	  (zero_extend:V8HI
	    (vec_select:V8QI
	      (match_operand:V16QI 1 "register_operand" "%f")
	      (parallel [(const_int 1) (const_int 3)
			 (const_int 5) (const_int 7)
			 (const_int 9) (const_int 11)
			 (const_int 13) (const_int 15)])))
	  (sign_extend:V8HI
	    (vec_select:V8QI
              (match_operand:V16QI 2 "register_operand" "f")
	      (parallel [(const_int 1) (const_int 3)
			 (const_int 5) (const_int 7)
			 (const_int 9) (const_int 11)
			 (const_int 13) (const_int 15)])))))]
  "ISA_HAS_LSX"
  "v<optab>wod.h.bu.b\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V8HI")])

(define_insn "lsx_vaddwev_q_d"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
		     (match_operand:V2DI 2 "register_operand" "f")]
		     UNSPEC_LSX_VADDWEV))]
  "ISA_HAS_LSX"
  "vaddwev.q.d\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vaddwev_q_du"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
		     (match_operand:V2DI 2 "register_operand" "f")]
		     UNSPEC_LSX_VADDWEV2))]
  "ISA_HAS_LSX"
  "vaddwev.q.du\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vaddwod_q_d"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
		     (match_operand:V2DI 2 "register_operand" "f")]
		     UNSPEC_LSX_VADDWOD))]
  "ISA_HAS_LSX"
  "vaddwod.q.d\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vaddwod_q_du"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
		     (match_operand:V2DI 2 "register_operand" "f")]
		     UNSPEC_LSX_VADDWOD2))]
  "ISA_HAS_LSX"
  "vaddwod.q.du\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vsubwev_q_d"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
		     (match_operand:V2DI 2 "register_operand" "f")]
		     UNSPEC_LSX_VSUBWEV))]
  "ISA_HAS_LSX"
  "vsubwev.q.d\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vsubwev_q_du"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
		     (match_operand:V2DI 2 "register_operand" "f")]
		     UNSPEC_LSX_VSUBWEV2))]
  "ISA_HAS_LSX"
  "vsubwev.q.du\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vsubwod_q_d"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
		     (match_operand:V2DI 2 "register_operand" "f")]
		     UNSPEC_LSX_VSUBWOD))]
  "ISA_HAS_LSX"
  "vsubwod.q.d\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vsubwod_q_du"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
		     (match_operand:V2DI 2 "register_operand" "f")]
		     UNSPEC_LSX_VSUBWOD2))]
  "ISA_HAS_LSX"
  "vsubwod.q.du\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vaddwev_q_du_d"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
		     (match_operand:V2DI 2 "register_operand" "f")]
		     UNSPEC_LSX_VADDWEV3))]
  "ISA_HAS_LSX"
  "vaddwev.q.du.d\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vaddwod_q_du_d"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
		     (match_operand:V2DI 2 "register_operand" "f")]
		     UNSPEC_LSX_VADDWOD3))]
  "ISA_HAS_LSX"
  "vaddwod.q.du.d\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vmulwev_q_du_d"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
		     (match_operand:V2DI 2 "register_operand" "f")]
		     UNSPEC_LSX_VMULWEV3))]
  "ISA_HAS_LSX"
  "vmulwev.q.du.d\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vmulwod_q_du_d"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
		     (match_operand:V2DI 2 "register_operand" "f")]
		     UNSPEC_LSX_VMULWOD3))]
  "ISA_HAS_LSX"
  "vmulwod.q.du.d\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vmulwev_q_d"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
		     (match_operand:V2DI 2 "register_operand" "f")]
		     UNSPEC_LSX_VMULWEV))]
  "ISA_HAS_LSX"
  "vmulwev.q.d\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vmulwev_q_du"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
		     (match_operand:V2DI 2 "register_operand" "f")]
		     UNSPEC_LSX_VMULWEV2))]
  "ISA_HAS_LSX"
  "vmulwev.q.du\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vmulwod_q_d"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
		     (match_operand:V2DI 2 "register_operand" "f")]
		     UNSPEC_LSX_VMULWOD))]
  "ISA_HAS_LSX"
  "vmulwod.q.d\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vmulwod_q_du"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
		     (match_operand:V2DI 2 "register_operand" "f")]
		     UNSPEC_LSX_VMULWOD2))]
  "ISA_HAS_LSX"
  "vmulwod.q.du\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vhaddw_q_d"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
		      (match_operand:V2DI 2 "register_operand" "f")]
		     UNSPEC_LSX_VHADDW_Q_D))]
  "ISA_HAS_LSX"
  "vhaddw.q.d\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vhaddw_qu_du"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
		      (match_operand:V2DI 2 "register_operand" "f")]
		     UNSPEC_LSX_VHADDW_QU_DU))]
  "ISA_HAS_LSX"
  "vhaddw.qu.du\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vhsubw_q_d"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
		      (match_operand:V2DI 2 "register_operand" "f")]
		     UNSPEC_LSX_VHSUBW_Q_D))]
  "ISA_HAS_LSX"
  "vhsubw.q.d\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vhsubw_qu_du"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
		      (match_operand:V2DI 2 "register_operand" "f")]
		     UNSPEC_LSX_VHSUBW_QU_DU))]
  "ISA_HAS_LSX"
  "vhsubw.qu.du\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vmaddwev_d_w<u>"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
   	(plus:V2DI
	  (match_operand:V2DI 1 "register_operand" "0")
	  (mult:V2DI
	    (any_extend:V2DI
	      (vec_select:V2SI
	        (match_operand:V4SI 2 "register_operand" "%f")
	        (parallel [(const_int 0) (const_int 2)])))
	    (any_extend:V2DI
	      (vec_select:V2SI
	        (match_operand:V4SI 3 "register_operand" "f")
	        (parallel [(const_int 0) (const_int 2)]))))))]
  "ISA_HAS_LSX"
  "vmaddwev.d.w<u>\t%w0,%w2,%w3"
  [(set_attr "type" "simd_fmadd")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vmaddwev_w_h<u>"
  [(set (match_operand:V4SI 0 "register_operand" "=f")
   	(plus:V4SI
	  (match_operand:V4SI 1 "register_operand" "0")
	  (mult:V4SI
	    (any_extend:V4SI
	      (vec_select:V4HI
	        (match_operand:V8HI 2 "register_operand" "%f")
	        (parallel [(const_int 0) (const_int 2)
	        	   (const_int 4) (const_int 6)])))
	    (any_extend:V4SI
	      (vec_select:V4HI
	        (match_operand:V8HI 3 "register_operand" "f")
	        (parallel [(const_int 0) (const_int 2)
			   (const_int 4) (const_int 6)]))))))]
  "ISA_HAS_LSX"
  "vmaddwev.w.h<u>\t%w0,%w2,%w3"
  [(set_attr "type" "simd_fmadd")
   (set_attr "mode" "V4SI")])

(define_insn "lsx_vmaddwev_h_b<u>"
  [(set (match_operand:V8HI 0 "register_operand" "=f")
	(plus:V8HI
	  (match_operand:V8HI 1 "register_operand" "0")
	  (mult:V8HI
	    (any_extend:V8HI
	      (vec_select:V8QI
	        (match_operand:V16QI 2 "register_operand" "%f")
	        (parallel [(const_int 0) (const_int 2)
			   (const_int 4) (const_int 6)
			   (const_int 8) (const_int 10)
			   (const_int 12) (const_int 14)])))
	    (any_extend:V8HI
	      (vec_select:V8QI
                (match_operand:V16QI 3 "register_operand" "f")
	        (parallel [(const_int 0) (const_int 2)
			   (const_int 4) (const_int 6)
			   (const_int 8) (const_int 10)
			   (const_int 12) (const_int 14)]))))))]
  "ISA_HAS_LSX"
  "vmaddwev.h.b<u>\t%w0,%w2,%w3"
  [(set_attr "type" "simd_fmadd")
   (set_attr "mode" "V8HI")])

(define_insn "lsx_vmaddwod_d_w<u>"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
   	(plus:V2DI
	  (match_operand:V2DI 1 "register_operand" "0")
	  (mult:V2DI
	    (any_extend:V2DI
	      (vec_select:V2SI
	        (match_operand:V4SI 2 "register_operand" "%f")
	        (parallel [(const_int 1) (const_int 3)])))
	    (any_extend:V2DI
	      (vec_select:V2SI
	        (match_operand:V4SI 3 "register_operand" "f")
	        (parallel [(const_int 1) (const_int 3)]))))))]
  "ISA_HAS_LSX"
  "vmaddwod.d.w<u>\t%w0,%w2,%w3"
  [(set_attr "type" "simd_fmadd")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vmaddwod_w_h<u>"
  [(set (match_operand:V4SI 0 "register_operand" "=f")
   	(plus:V4SI
	  (match_operand:V4SI 1 "register_operand" "0")
	  (mult:V4SI
	    (any_extend:V4SI
	      (vec_select:V4HI
	        (match_operand:V8HI 2 "register_operand" "%f")
	        (parallel [(const_int 1) (const_int 3)
	        	   (const_int 5) (const_int 7)])))
	    (any_extend:V4SI
	      (vec_select:V4HI
	        (match_operand:V8HI 3 "register_operand" "f")
	        (parallel [(const_int 1) (const_int 3)
			   (const_int 5) (const_int 7)]))))))]
  "ISA_HAS_LSX"
  "vmaddwod.w.h<u>\t%w0,%w2,%w3"
  [(set_attr "type" "simd_fmadd")
   (set_attr "mode" "V4SI")])

(define_insn "lsx_vmaddwod_h_b<u>"
  [(set (match_operand:V8HI 0 "register_operand" "=f")
	(plus:V8HI
	  (match_operand:V8HI 1 "register_operand" "0")
	  (mult:V8HI
	    (any_extend:V8HI
	      (vec_select:V8QI
	        (match_operand:V16QI 2 "register_operand" "%f")
	        (parallel [(const_int 1) (const_int 3)
			   (const_int 5) (const_int 7)
			   (const_int 9) (const_int 11)
			   (const_int 13) (const_int 15)])))
	    (any_extend:V8HI
	      (vec_select:V8QI
                (match_operand:V16QI 3 "register_operand" "f")
	        (parallel [(const_int 1) (const_int 3)
			   (const_int 5) (const_int 7)
			   (const_int 9) (const_int 11)
			   (const_int 13) (const_int 15)]))))))]
  "ISA_HAS_LSX"
  "vmaddwod.h.b<u>\t%w0,%w2,%w3"
  [(set_attr "type" "simd_fmadd")
   (set_attr "mode" "V8HI")])

(define_insn "lsx_vmaddwev_d_wu_w"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
   	(plus:V2DI
	  (match_operand:V2DI 1 "register_operand" "0")
	  (mult:V2DI
	    (zero_extend:V2DI
	      (vec_select:V2SI
	        (match_operand:V4SI 2 "register_operand" "%f")
	        (parallel [(const_int 0) (const_int 2)])))
	    (sign_extend:V2DI
	      (vec_select:V2SI
	        (match_operand:V4SI 3 "register_operand" "f")
	        (parallel [(const_int 0) (const_int 2)]))))))]
  "ISA_HAS_LSX"
  "vmaddwev.d.wu.w\t%w0,%w2,%w3"
  [(set_attr "type" "simd_fmadd")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vmaddwev_w_hu_h"
  [(set (match_operand:V4SI 0 "register_operand" "=f")
   	(plus:V4SI
	  (match_operand:V4SI 1 "register_operand" "0")
	  (mult:V4SI
	    (zero_extend:V4SI
	      (vec_select:V4HI
	        (match_operand:V8HI 2 "register_operand" "%f")
	        (parallel [(const_int 0) (const_int 2)
	        	   (const_int 4) (const_int 6)])))
	    (sign_extend:V4SI
	      (vec_select:V4HI
	        (match_operand:V8HI 3 "register_operand" "f")
	        (parallel [(const_int 0) (const_int 2)
			   (const_int 4) (const_int 6)]))))))]
  "ISA_HAS_LSX"
  "vmaddwev.w.hu.h\t%w0,%w2,%w3"
  [(set_attr "type" "simd_fmadd")
   (set_attr "mode" "V4SI")])

(define_insn "lsx_vmaddwev_h_bu_b"
  [(set (match_operand:V8HI 0 "register_operand" "=f")
	(plus:V8HI
	  (match_operand:V8HI 1 "register_operand" "0")
	  (mult:V8HI
	    (zero_extend:V8HI
	      (vec_select:V8QI
	        (match_operand:V16QI 2 "register_operand" "%f")
	        (parallel [(const_int 0) (const_int 2)
			   (const_int 4) (const_int 6)
			   (const_int 8) (const_int 10)
			   (const_int 12) (const_int 14)])))
	    (sign_extend:V8HI
	      (vec_select:V8QI
                (match_operand:V16QI 3 "register_operand" "f")
	        (parallel [(const_int 0) (const_int 2)
			   (const_int 4) (const_int 6)
			   (const_int 8) (const_int 10)
			   (const_int 12) (const_int 14)]))))))]
  "ISA_HAS_LSX"
  "vmaddwev.h.bu.b\t%w0,%w2,%w3"
  [(set_attr "type" "simd_fmadd")
   (set_attr "mode" "V8HI")])

(define_insn "lsx_vmaddwod_d_wu_w"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
   	(plus:V2DI
	  (match_operand:V2DI 1 "register_operand" "0")
	  (mult:V2DI
	    (zero_extend:V2DI
	      (vec_select:V2SI
	        (match_operand:V4SI 2 "register_operand" "%f")
	        (parallel [(const_int 1) (const_int 3)])))
	    (sign_extend:V2DI
	      (vec_select:V2SI
	        (match_operand:V4SI 3 "register_operand" "f")
	        (parallel [(const_int 1) (const_int 3)]))))))]
  "ISA_HAS_LSX"
  "vmaddwod.d.wu.w\t%w0,%w2,%w3"
  [(set_attr "type" "simd_fmadd")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vmaddwod_w_hu_h"
  [(set (match_operand:V4SI 0 "register_operand" "=f")
   	(plus:V4SI
	  (match_operand:V4SI 1 "register_operand" "0")
	  (mult:V4SI
	    (zero_extend:V4SI
	      (vec_select:V4HI
	        (match_operand:V8HI 2 "register_operand" "%f")
	        (parallel [(const_int 1) (const_int 3)
	        	   (const_int 5) (const_int 7)])))
	    (sign_extend:V4SI
	      (vec_select:V4HI
	        (match_operand:V8HI 3 "register_operand" "f")
	        (parallel [(const_int 1) (const_int 3)
			   (const_int 5) (const_int 7)]))))))]
  "ISA_HAS_LSX"
  "vmaddwod.w.hu.h\t%w0,%w2,%w3"
  [(set_attr "type" "simd_fmadd")
   (set_attr "mode" "V4SI")])

(define_insn "lsx_vmaddwod_h_bu_b"
  [(set (match_operand:V8HI 0 "register_operand" "=f")
	(plus:V8HI
	  (match_operand:V8HI 1 "register_operand" "0")
	  (mult:V8HI
	    (zero_extend:V8HI
	      (vec_select:V8QI
	        (match_operand:V16QI 2 "register_operand" "%f")
	        (parallel [(const_int 1) (const_int 3)
			   (const_int 5) (const_int 7)
			   (const_int 9) (const_int 11)
			   (const_int 13) (const_int 15)])))
	    (sign_extend:V8HI
	      (vec_select:V8QI
                (match_operand:V16QI 3 "register_operand" "f")
	        (parallel [(const_int 1) (const_int 3)
			   (const_int 5) (const_int 7)
			   (const_int 9) (const_int 11)
			   (const_int 13) (const_int 15)]))))))]
  "ISA_HAS_LSX"
  "vmaddwod.h.bu.b\t%w0,%w2,%w3"
  [(set_attr "type" "simd_fmadd")
   (set_attr "mode" "V8HI")])

(define_insn "lsx_vmaddwev_q_d"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "0")
		     (match_operand:V2DI 2 "register_operand" "f")
		     (match_operand:V2DI 3 "register_operand" "f")]
		     UNSPEC_LSX_VMADDWEV))]
  "ISA_HAS_LSX"
  "vmaddwev.q.d\t%w0,%w2,%w3"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vmaddwod_q_d"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "0")
		     (match_operand:V2DI 2 "register_operand" "f")
		     (match_operand:V2DI 3 "register_operand" "f")]
		     UNSPEC_LSX_VMADDWOD))]
  "ISA_HAS_LSX"
  "vmaddwod.q.d\t%w0,%w2,%w3"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vmaddwev_q_du"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "0")
		     (match_operand:V2DI 2 "register_operand" "f")
		     (match_operand:V2DI 3 "register_operand" "f")]
		     UNSPEC_LSX_VMADDWEV2))]
  "ISA_HAS_LSX"
  "vmaddwev.q.du\t%w0,%w2,%w3"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vmaddwod_q_du"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "0")
		     (match_operand:V2DI 2 "register_operand" "f")
		     (match_operand:V2DI 3 "register_operand" "f")]
		     UNSPEC_LSX_VMADDWOD2))]
  "ISA_HAS_LSX"
  "vmaddwod.q.du\t%w0,%w2,%w3"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vmaddwev_q_du_d"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "0")
		     (match_operand:V2DI 2 "register_operand" "f")
		     (match_operand:V2DI 3 "register_operand" "f")]
		     UNSPEC_LSX_VMADDWEV3))]
  "ISA_HAS_LSX"
  "vmaddwev.q.du.d\t%w0,%w2,%w3"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vmaddwod_q_du_d"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "0")
		     (match_operand:V2DI 2 "register_operand" "f")
		     (match_operand:V2DI 3 "register_operand" "f")]
		     UNSPEC_LSX_VMADDWOD3))]
  "ISA_HAS_LSX"
  "vmaddwod.q.du.d\t%w0,%w2,%w3"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vrotr_<lsxfmt>"
  [(set (match_operand:ILSX 0 "register_operand" "=f")
        (unspec:ILSX [(match_operand:ILSX 1 "register_operand" "f")
                     (match_operand:ILSX 2 "register_operand" "f")]
                     UNSPEC_LSX_VROTR))]
  "ISA_HAS_LSX"
  "vrotr.<lsxfmt>\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "<MODE>")])

(define_insn "lsx_vadd_q"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
        (unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
                     (match_operand:V2DI 2 "register_operand" "f")]
                     UNSPEC_LSX_VADD_Q))]
  "ISA_HAS_LSX"
  "vadd.q\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vsub_q"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
        (unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")
                     (match_operand:V2DI 2 "register_operand" "f")]
                     UNSPEC_LSX_VSUB_Q))]
  "ISA_HAS_LSX"
  "vsub.q\t%w0,%w1,%w2"
  [(set_attr "type" "simd_int_arith")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vmskgez_b"
  [(set (match_operand:V16QI 0 "register_operand" "=f")
	(unspec:V16QI [(match_operand:V16QI 1 "register_operand" "f")]
		     UNSPEC_LSX_VMSKGEZ))]
  "ISA_HAS_LSX"
  "vmskgez.b\t%w0,%w1"
  [(set_attr "type" "simd_bit")
   (set_attr "mode" "V16QI")])

(define_insn "lsx_vmsknz_b"
  [(set (match_operand:V16QI 0 "register_operand" "=f")
	(unspec:V16QI [(match_operand:V16QI 1 "register_operand" "f")]
		     UNSPEC_LSX_VMSKNZ))]
  "ISA_HAS_LSX"
  "vmsknz.b\t%w0,%w1"
  [(set_attr "type" "simd_bit")
   (set_attr "mode" "V16QI")])

(define_insn "lsx_vexth_h<u>_b<u>"
  [(set (match_operand:V8HI 0 "register_operand" "=f")
	(any_extend:V8HI
	    (vec_select:V8QI
	        (match_operand:V16QI 1 "register_operand" "f")
	        (parallel [(const_int 8) (const_int 9)
	        	   (const_int 10) (const_int 11)
	        	   (const_int 12) (const_int 13)
			   (const_int 14) (const_int 15)]))))]
  "ISA_HAS_LSX"
  "vexth.h<u>.b<u>\t%w0,%w1"
  [(set_attr "type" "simd_fcvt")
   (set_attr "mode" "V8HI")])

(define_insn "lsx_vexth_w<u>_h<u>"
  [(set (match_operand:V4SI 0 "register_operand" "=f")
	(any_extend:V4SI
	    (vec_select:V4HI
	        (match_operand:V8HI 1 "register_operand" "f")
	        (parallel [(const_int 4) (const_int 5)
			   (const_int 6) (const_int 7)]))))]
  "ISA_HAS_LSX"
  "vexth.w<u>.h<u>\t%w0,%w1"
  [(set_attr "type" "simd_fcvt")
   (set_attr "mode" "V4SI")])

(define_insn "lsx_vexth_d<u>_w<u>"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(any_extend:V2DI
	    (vec_select:V2SI
	        (match_operand:V4SI 1 "register_operand" "f")
	        (parallel [(const_int 2) (const_int 3)]))))]
  "ISA_HAS_LSX"
  "vexth.d<u>.w<u>\t%w0,%w1"
  [(set_attr "type" "simd_fcvt")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vexth_q_d"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")]
		   UNSPEC_LSX_VEXTH_Q_D))]
  "ISA_HAS_LSX"
  "vexth.q.d\t%w0,%w1"
  [(set_attr "type" "simd_fcvt")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vexth_qu_du"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")]
		   UNSPEC_LSX_VEXTH_QU_DU))]
  "ISA_HAS_LSX"
  "vexth.qu.du\t%w0,%w1"
  [(set_attr "type" "simd_fcvt")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vrotri_<lsxfmt>"
  [(set (match_operand:ILSX 0 "register_operand" "=f")
	(unspec:ILSX [(match_operand:ILSX 1 "register_operand" "f")
		      (match_operand 2 "const_<bitimm>_operand" "")]
		   UNSPEC_LSX_VROTRI))]
  "ISA_HAS_LSX"
  "vrotri.<lsxfmt>\t%w0,%w1,%2"
  [(set_attr "type" "simd_shf")
   (set_attr "mode" "<MODE>")])

(define_insn "lsx_vextl_q_d"
  [(set (match_operand:V2DI 0 "register_operand" "=f")
	(unspec:V2DI [(match_operand:V2DI 1 "register_operand" "f")]
		   UNSPEC_LSX_VEXTL_Q_D))]
  "ISA_HAS_LSX"
  "vextl.q.d\t%w0,%w1"
  [(set_attr "type" "simd_fcvt")
   (set_attr "mode" "V2DI")])

(define_insn "lsx_vsrlni_<lsxfmt>_<dlsxfmt>"
  [(set (match_operand:ILSX 0 "register_operand" "=f")
	(unspec:ILSX [(match_operand:ILSX 1 "register_operand" "0")
                          (match_operand:ILSX 2 "register_operand" "f")
			  (match_operand 3 "const_uimm8_operand" "")]
			 UNSPEC_LSX_VSRLNI))]
  "ISA_HAS_LSX"
  "vsrlni.<lsxfmt>.<dlsxfmt>\t%w0,%w2,%3"
  [(set_attr "type" "simd_shift")
   (set_attr "mode" "<MODE>")])

(define_insn "lsx_vsrlrni_<lsxfmt>_<dlsxfmt>"
  [(set (match_operand:ILSX 0 "register_operand" "=f")
	(unspec:ILSX [(match_operand:ILSX 1 "register_operand" "0")
                          (match_operand:ILSX 2 "register_operand" "f")
			  (match_operand 3 "const_uimm8_operand" "")]
			 UNSPEC_LSX_VSRLRNI))]
  "ISA_HAS_LSX"
  "vsrlrni.<lsxfmt>.<dlsxfmt>\t%w0,%w2,%3"
  [(set_attr "type" "simd_shift")
   (set_attr "mode" "<MODE>")])

(define_insn "lsx_vssrlni_<lsxfmt>_<dlsxfmt>"
  [(set (match_operand:ILSX 0 "register_operand" "=f")
	(unspec:ILSX [(match_operand:ILSX 1 "register_operand" "0")
                          (match_operand:ILSX 2 "register_operand" "f")
			  (match_operand 3 "const_uimm8_operand" "")]
			 UNSPEC_LSX_VSSRLNI))]
  "ISA_HAS_LSX"
  "vssrlni.<lsxfmt>.<dlsxfmt>\t%w0,%w2,%3"
  [(set_attr "type" "simd_shift")
   (set_attr "mode" "<MODE>")])

(define_insn "lsx_vssrlni_<lsxfmt_u>_<dlsxfmt>"
  [(set (match_operand:ILSX 0 "register_operand" "=f")
	(unspec:ILSX [(match_operand:ILSX 1 "register_operand" "0")
                          (match_operand:ILSX 2 "register_operand" "f")
			  (match_operand 3 "const_uimm8_operand" "")]
			 UNSPEC_LSX_VSSRLNI2))]
  "ISA_HAS_LSX"
  "vssrlni.<lsxfmt_u>.<dlsxfmt>\t%w0,%w2,%3"
  [(set_attr "type" "simd_shift")
   (set_attr "mode" "<MODE>")])

(define_insn "lsx_vssrlrni_<lsxfmt>_<dlsxfmt>"
  [(set (match_operand:ILSX 0 "register_operand" "=f")
	(unspec:ILSX [(match_operand:ILSX 1 "register_operand" "0")
                          (match_operand:ILSX 2 "register_operand" "f")
			  (match_operand 3 "const_uimm8_operand" "")]
			 UNSPEC_LSX_VSSRLRNI))]
  "ISA_HAS_LSX"
  "vssrlrni.<lsxfmt>.<dlsxfmt>\t%w0,%w2,%3"
  [(set_attr "type" "simd_shift")
   (set_attr "mode" "<MODE>")])

(define_insn "lsx_vssrlrni_<lsxfmt_u>_<dlsxfmt>"
  [(set (match_operand:ILSX 0 "register_operand" "=f")
	(unspec:ILSX [(match_operand:ILSX 1 "register_operand" "0")
                          (match_operand:ILSX 2 "register_operand" "f")
			  (match_operand 3 "const_uimm8_operand" "")]
			 UNSPEC_LSX_VSSRLRNI2))]
  "ISA_HAS_LSX"
  "vssrlrni.<lsxfmt_u>.<dlsxfmt>\t%w0,%w2,%3"
  [(set_attr "type" "simd_shift")
   (set_attr "mode" "<MODE>")])

(define_insn "lsx_vsrani_<lsxfmt>_<dlsxfmt>"
  [(set (match_operand:ILSX 0 "register_operand" "=f")
	(unspec:ILSX [(match_operand:ILSX 1 "register_operand" "0")
                          (match_operand:ILSX 2 "register_operand" "f")
			  (match_operand 3 "const_uimm8_operand" "")]
			 UNSPEC_LSX_VSRANI))]
  "ISA_HAS_LSX"
  "vsrani.<lsxfmt>.<dlsxfmt>\t%w0,%w2,%3"
  [(set_attr "type" "simd_shift")
   (set_attr "mode" "<MODE>")])

(define_insn "lsx_vsrarni_<lsxfmt>_<dlsxfmt>"
  [(set (match_operand:ILSX 0 "register_operand" "=f")
	(unspec:ILSX [(match_operand:ILSX 1 "register_operand" "0")
                          (match_operand:ILSX 2 "register_operand" "f")
			  (match_operand 3 "const_uimm8_operand" "")]
			 UNSPEC_LSX_VSRARNI))]
  "ISA_HAS_LSX"
  "vsrarni.<lsxfmt>.<dlsxfmt>\t%w0,%w2,%3"
  [(set_attr "type" "simd_shift")
   (set_attr "mode" "<MODE>")])

(define_insn "lsx_vssrani_<lsxfmt>_<dlsxfmt>"
  [(set (match_operand:ILSX 0 "register_operand" "=f")
	(unspec:ILSX [(match_operand:ILSX 1 "register_operand" "0")
                          (match_operand:ILSX 2 "register_operand" "f")
			  (match_operand 3 "const_uimm8_operand" "")]
			 UNSPEC_LSX_VSSRANI))]
  "ISA_HAS_LSX"
  "vssrani.<lsxfmt>.<dlsxfmt>\t%w0,%w2,%3"
  [(set_attr "type" "simd_shift")
   (set_attr "mode" "<MODE>")])

(define_insn "lsx_vssrani_<lsxfmt_u>_<dlsxfmt>"
  [(set (match_operand:ILSX 0 "register_operand" "=f")
	(unspec:ILSX [(match_operand:ILSX 1 "register_operand" "0")
                          (match_operand:ILSX 2 "register_operand" "f")
			  (match_operand 3 "const_uimm8_operand" "")]
			 UNSPEC_LSX_VSSRANI2))]
  "ISA_HAS_LSX"
  "vssrani.<lsxfmt_u>.<dlsxfmt>\t%w0,%w2,%3"
  [(set_attr "type" "simd_shift")
   (set_attr "mode" "<MODE>")])

(define_insn "lsx_vssrarni_<lsxfmt>_<dlsxfmt>"
  [(set (match_operand:ILSX 0 "register_operand" "=f")
	(unspec:ILSX [(match_operand:ILSX 1 "register_operand" "0")
                          (match_operand:ILSX 2 "register_operand" "f")
			  (match_operand 3 "const_uimm8_operand" "")]
			 UNSPEC_LSX_VSSRARNI))]
  "ISA_HAS_LSX"
  "vssrarni.<lsxfmt>.<dlsxfmt>\t%w0,%w2,%3"
  [(set_attr "type" "simd_shift")
   (set_attr "mode" "<MODE>")])

(define_insn "lsx_vssrarni_<lsxfmt_u>_<dlsxfmt>"
  [(set (match_operand:ILSX 0 "register_operand" "=f")
	(unspec:ILSX [(match_operand:ILSX 1 "register_operand" "0")
                          (match_operand:ILSX 2 "register_operand" "f")
			  (match_operand 3 "const_uimm8_operand" "")]
			 UNSPEC_LSX_VSSRARNI2))]
  "ISA_HAS_LSX"
  "vssrarni.<lsxfmt_u>.<dlsxfmt>\t%w0,%w2,%3"
  [(set_attr "type" "simd_shift")
   (set_attr "mode" "<MODE>")])

(define_insn "lsx_vpermi_w"
  [(set (match_operand:V4SI 0 "register_operand" "=f")
	(unspec:V4SI [(match_operand:V4SI 1 "register_operand" "0")
               (match_operand:V4SI 2 "register_operand" "f")
	     	       (match_operand 3 "const_uimm8_operand" "")]
		     UNSPEC_LSX_VPERMI))]
  "ISA_HAS_LSX"
  "vpermi.w\t%w0,%w2,%3"
  [(set_attr "type" "simd_bit")
   (set_attr "mode" "V4SI")])

